/*
 * SmartcardioNFC - a library providing a smartcardio implementation for NFC on Android
 *
 * Copyright (C) 2017 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH <info@gematik.de> +49 30 40041-0
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.nfcCard;

import android.app.Activity;
import android.nfc.Tag;
import android.nfc.tech.IsoDep;
import android.util.Log;

import java.io.IOException;
import java.util.List;

import javax.smartcardio.ATR;
import javax.smartcardio.Card;
import javax.smartcardio.CardException;
import javax.smartcardio.CardPermission;
import javax.smartcardio.CardTerminals;
import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;
import javax.smartcardio.TerminalFactory;

import de.gematik.security.Utilities.Utilities;
import de.gematik.security.iso7816.Constants;

/**
 * A Nfc Smart Card with which a connection has been established. Card objects
 * are obtained by calling {@link NfcCardTerminal#connect NfcCardTerminal.connect()}.
 *
 * @see NfcCardTerminal
 */
public class NfcCard extends Card {

    private static final String TAG = NfcCard.class.getName();
    private IsoDep isoDep;
    private boolean isCardConnected = false;
    private NfcCardChannel basicChannel = null;

    /**
     * Convenience method that waits indefinitely for an inserted card and creates a connection to it.
     *
     * @param activity activity of Android app
     * @param protocol the protocol to use ("T=0", "T=1", or "T=CL"), or "*" to
     *   connect using any available protocol.
     * @return an instance of NfcCard
     * @throws CardException if any card I/O error occurred
     */
    public static NfcCard waitForCardAndConnect(Activity activity, String protocol) throws CardException {
        NfcCardTerminals nfcCardTerminals = (NfcCardTerminals) TerminalFactory.getDefault(activity).terminals();
//        NfcCardTerminals nfcCardTerminals = (NfcCardTerminals) TerminalFactory.getInstance("TerminalFactory.GematikNfcHealthCard", activity, "NfcHealthCardProvider").terminals();
        while(true) {
            nfcCardTerminals.waitForChange();
            List<NfcCardTerminal> list = nfcCardTerminals.list(CardTerminals.State.ALL);
            for (int i = 0; i < list.size(); i++) {
                NfcCardTerminal item = list.get(i);
                if (item != null && item.isCardPresent()) {
                    return item.connect(protocol);
                }
            }
        }
    }

    /**
     * Constructs a new Card object.
     * <p/>
     * <p>This constructor is called by subclasses only. Application should
     * call the {@linkplain NfcCardTerminal#connect NfcCardTerminal.connect()}
     * method to obtain a Card
     * object.
     */
    protected NfcCard() {
        // empty
    }

    protected NfcCard(Tag tag) {
        isoDep = IsoDep.get(tag);
        if (isoDep != null) {
            try {
                Log.d(TAG, "try isoDep.connect()");
                isoDep.connect();
                Log.d(TAG, "isoDep.getMaxTransceiveLength(): " + isoDep.getMaxTransceiveLength());
                Log.d(TAG, "original value for isoDep.getTimeout(): " + isoDep.getTimeout());
                isoDep.setTimeout(3000);
                Log.d(TAG, "new value for isoDep.getTimeout(): " + isoDep.getTimeout());
                basicChannel = new NfcCardChannel(this, false);
            } catch (IOException e) {
                Log.e(TAG, "Connect: IsoDep unsuccessful connect", e);
            } catch (CardException e) {
                Log.e(TAG, "Connect: Cannot establish basic channel", e);
            }
        }
    }

    protected boolean isCardPresent() {
        boolean result = isoDep.isConnected();
//        Log.d(TAG, "isCardPresent() = " + result);
//        Log.d(TAG, "isCardConnected = " + isCardConnected);
        return result;
    }

    protected void connect(){
        isCardConnected = true;
    }

    protected ResponseAPDU transceive(CommandAPDU commandAPDU) throws CardException {
        if (isoDep == null || !isoDep.isConnected() || !isCardConnected) {
            throw new CardException("Card is not connected (anymore)");
        }
        Log.d(TAG, "SEND: " + Utilities.ByteArrayToSpacedHexString(commandAPDU.getBytes()));
        ResponseAPDU responseAPDU;
        if(commandAPDU.getBytes().length>6&&commandAPDU.getBytes()[4]==0&&!isoDep.isExtendedLengthApduSupported()){
            throw new CardException("Extended length APDU not supported!");
        }
        try {
            responseAPDU = new ResponseAPDU(isoDep.transceive(commandAPDU.getBytes()));
        } catch (IllegalArgumentException e) {
            throw new CardException("Transceive APDU IllegalArgumentException", e);
        } catch (IOException e) {
            throw new CardException("Transceive APDU IOException", e);
        }
        Log.d(TAG, "RECEIVED: " + Utilities.ByteArrayToSpacedHexString(responseAPDU.getBytes()));
        return responseAPDU;
    }

    /**
     * Returns the ATR of this card.
     *
     * @return the ATR of this card.
     */
    public ATR getATR() {
        if (isoDep != null) {
            byte[] data = isoDep.getHistoricalBytes();
            Log.d(TAG, "data: " + data);
            if(data==null){
                data = isoDep.getHiLayerResponse();
                Log.d(TAG, "data: " + data);
            }
            if(data!=null){
                return new ATR(data);
            }
        }
        return new ATR(new byte[0]);
    }

    /**
     * Returns the protocol in use for this card.
     *
     * @return the protocol in use for this card, for example "T=0" or "T=1"
     */
    public String getProtocol() {
        return "T=CL";
    }

    /**
     * Returns the CardChannel for the basic logical channel. The basic
     * logical channel has a channel number of 0.
     *
     * @throws SecurityException     if a SecurityManager exists and the
     *                               caller does not have the required
     *                               {@linkplain CardPermission permission}
     * @throws IllegalStateException if this card object has been disposed of
     *                               via the {@linkplain #disconnect disconnect()} method
     */
    public NfcCardChannel getBasicChannel() {
        return basicChannel;
    }

    /**
     * Opens a new logical channel to the card and returns it. The channel is
     * opened by issuing a <code>MANAGE CHANNEL</code> command that should use
     * the format <code>[00 70 00 00 01]</code>.
     *
     * @throws SecurityException     if a SecurityManager exists and the
     *                               caller does not have the required
     *                               {@linkplain CardPermission permission}
     * @throws CardException         is a new logical channel could not be opened
     * @throws IllegalStateException if this card object has been disposed of
     *                               via the {@linkplain #disconnect disconnect()} method
     */
    public NfcCardChannel openLogicalChannel() throws CardException {
        return new NfcCardChannel(this, true);
    }

    /**
     * Requests exclusive access to this card.
     * <p/>
     * <p>Once a thread has invoked <code>beginExclusive</code>, only this
     * thread is allowed to communicate with this card until it calls
     * <code>endExclusive</code>. Other threads attempting communication
     * will receive a CardException.
     * <p/>
     * <p>Applications have to ensure that exclusive access is correctly
     * released. This can be achieved by executing
     * the <code>beginExclusive()</code> and <code>endExclusive</code> calls
     * in a <code>try ... finally</code> block.
     *
     * @throws SecurityException     if a SecurityManager exists and the
     *                               caller does not have the required
     *                               {@linkplain CardPermission permission}
     * @throws CardException         if exclusive access has already been set
     *                               or if exclusive access could not be established
     * @throws IllegalStateException if this card object has been disposed of
     *                               via the {@linkplain #disconnect disconnect()} method
     */
    public void beginExclusive() throws CardException {
        //not supported
        throw new CardException("operation not supported");
    }

    /**
     * Releases the exclusive access previously established using
     * <code>beginExclusive</code>.
     *
     * @throws SecurityException     if a SecurityManager exists and the
     *                               caller does not have the required
     *                               {@linkplain CardPermission permission}
     * @throws IllegalStateException if the active Thread does not currently have
     *                               exclusive access to this card or
     *                               if this card object has been disposed of
     *                               via the {@linkplain #disconnect disconnect()} method
     * @throws CardException         if the operation failed
     */
    public void endExclusive() throws CardException {
        //not supported
        throw new CardException("operation not supported");
    }

    /**
     * Transmits a control command to the terminal device.
     * <p/>
     * <p>This can be used to, for example, control terminal functions like
     * a built-in PIN pad or biometrics.
     *
     * @param controlCode the control code of the command
     * @param command     the command data
     * @throws SecurityException     if a SecurityManager exists and the
     *                               caller does not have the required
     *                               {@linkplain CardPermission permission}
     * @throws NullPointerException  if command is null
     * @throws CardException         if the card operation failed
     * @throws IllegalStateException if this card object has been disposed of
     *                               via the {@linkplain #disconnect disconnect()} method
     */
    public byte[] transmitControlCommand(int controlCode, byte[] command) throws CardException {
        //not supported
        throw new CardException("control commands not supported");
    }

    /**
     * Disconnects the connection with this card. After this method returns,
     * calling methods on this object or in CardChannels associated with this
     * object that require interaction with the card will raise an
     * IllegalStateException.
     *
     * @param reset whether to reset the card after disconnecting.
     * @throws CardException     if the card operation failed
     * @throws SecurityException if a SecurityManager exists and the
     *                           caller does not have the required
     *                           {@linkplain CardPermission permission}
     */
    public void disconnect(boolean reset) throws CardException {
        CommandAPDU commandAPDU = new CommandAPDU(0x00, 0x70, 0x40, 0x01);
        ResponseAPDU responseAPDU;
        responseAPDU = transceive(commandAPDU);
        isCardConnected = false;
        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            throw new CardException("RESET CHANNELs unexpected response: " + responseAPDU.getSW());
        }
        if(reset) {
            try {
                Tag tag = isoDep.getTag();
                isoDep.close();
                isoDep = IsoDep.get(tag);
                isoDep.connect();
                isoDep.setTimeout(3000);
            } catch (IOException e) {
                throw new CardException("unsuccessful reset of isodep connection", e);
            }
        }
    }

    /**
     * @return true if this NfcCard supports ExtendedLengthApdus, false if not
     */
    public boolean isExtendedLengthApduSupported(){
        if(isoDep!=null){
            return isoDep.isExtendedLengthApduSupported();
        }else{
            return false;
        }
    }

}
