/*
 * SmartcardioNFC - a library providing a smartcardio implementation for NFC on Android
 *
 * Copyright (C) 2017 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH <info@gematik.de> +49 30 40041-0
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.nfcCard;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

import java.lang.ref.WeakReference;
import java.nio.ByteBuffer;
import java.util.Properties;

import javax.smartcardio.Card;
import javax.smartcardio.CardChannel;
import javax.smartcardio.CardException;
import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;

import de.gematik.security.Utilities.Utilities;
import de.gematik.security.iso7816.Constants;
import de.gematik.security.tr03110.SecureMessaging;
import de.gematik.security.tr03110.SecureMessagingException;

/**
 * A logical channel connection to a Nfc Smart Card. It is used to exchange APDUs
 * with a Smart Card using IsoDep.
 * A NfcCardChannel object can be obtained by calling the method
 * {@linkplain NfcCard#getBasicChannel} or {@linkplain NfcCard#openLogicalChannel}.
 *
 * @see Card
 * @see CommandAPDU
 * @see ResponseAPDU
 */
public class NfcCardChannel extends CardChannel {

    private static final int PRIVATE_KEY_MODULUS_LENGTH = 2048;

    private static final String TAG = NfcCardChannel.class.getName();
    private int channelNumber;
    private NfcCard nfcCard;
    private SecureMessaging secureMessaging = null;
    private WeakReference<Handler> weakReferenceHandler = null;
    private int what;
    private Status status = new Status();

    public static class Status {
        public int totalNumberOfReceivedBytes = 0;
        public int totalNumberOfSentBytes = 0;
    };

    protected NfcCardChannel() {
        throw new IllegalArgumentException("Card object required to create channel");
    }

    /**
     * Constructs a new CardChannel object.
     * <p/>
     * <p>This constructor is called by subclasses only. Application should
     * call the {@linkplain Card#getBasicChannel} and
     * {@linkplain Card#openLogicalChannel} methods to obtain a CardChannel
     * object.
     *
     * @param nfcCard   card object which requests to open the channel
     * @param isLogical indicates type of channel (false: basic channel, true logical channel)
     * @throws CardException if this MANAGE CHANNEL command return a error
     */
    protected NfcCardChannel(NfcCard nfcCard, boolean isLogical) throws CardException {
        this.nfcCard = nfcCard;
        if (isLogical) {
            CommandAPDU commandAPDU = new CommandAPDU(0x00, 0x70, 0x00, 0x00, 1);
            ResponseAPDU responseAPDU;
            responseAPDU = nfcCard.transceive(commandAPDU);
            if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
                throw new CardException("OPEN CHANNEL unexpected response: " + responseAPDU.getSW());
            }
            channelNumber = responseAPDU.getData()[0];
        } else {
            channelNumber = 0;
        }
    }

    /*-------------------------- Methods that implement abstract class CardChannel ------------------------*/

    @Override
    public Card getCard() {
        return nfcCard;
    }

    @Override
    public int getChannelNumber() {
        return channelNumber;
    }

    @Override
    public ResponseAPDU transmit(CommandAPDU command) throws CardException {
        if (command.getINS() == 0x70) {
            throw new IllegalArgumentException("MANAGE CHANNEL not allowed");
        }
        byte[] commandBytes = command.getBytes();
        commandBytes[0] = adjustChannel(commandBytes[0]);
        command = new CommandAPDU(commandBytes);
        ResponseAPDU response;

        // TODO find a better way to pass SecureMessaging instance between classes HealthcardContext and NfcCardChannel
        // (the problem is that HealthcardContext knows only CardChannel and not NfcCardChannel)
        Properties props = System.getProperties();
        secureMessaging = (SecureMessaging) props.get("secureMessaging");

        if (secureMessaging != null) {
            try {
                Log.d(TAG, "WRAP: " + Utilities.ByteArrayToSpacedHexString(command.getBytes()));
                command = secureMessaging.wrap(command);
                status.totalNumberOfSentBytes+=command.getBytes().length;
                response = secureMessaging.unwrap(nfcCard.transceive(command));
                status.totalNumberOfReceivedBytes+=response.getBytes().length;
                Log.d(TAG, "UNWRAPPED: " + Utilities.ByteArrayToSpacedHexString(response.getBytes()));
            } catch (SecureMessagingException e) {
                throw new CardException(TAG, e);
            }
        } else {
            status.totalNumberOfSentBytes+=command.getBytes().length;
            response = nfcCard.transceive(command);
            status.totalNumberOfReceivedBytes+=response.getBytes().length;
        }
        if(weakReferenceHandler!=null){
            Handler handler = weakReferenceHandler.get();
            if(handler!=null){
                Message msg = handler.obtainMessage(what, status);
                handler.sendMessage(msg);
            }
        }
        return response;
    }

    @Override
    public int transmit(ByteBuffer command, ByteBuffer response) throws CardException {
        byte[] commandBytes = new byte[command.remaining()];
        command.get(commandBytes, command.position(), command.remaining());
        CommandAPDU commandAPDU = new CommandAPDU(commandBytes);
        byte[] responseBytes = transmit(commandAPDU).getBytes();
        response.put(responseBytes, response.position(), responseBytes.length);
        return responseBytes.length;
    }

    @Override
    public void close() throws CardException {
        CommandAPDU commandAPDU = new CommandAPDU(adjustChannel((byte) 0x00), 0x70, 0x80, 0x00);
        ResponseAPDU responseAPDU;
        responseAPDU = nfcCard.transceive(commandAPDU);
        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            throw new CardException("CLOSE CHANNEL unexpected response: " + responseAPDU.getSW());
        }
    }

    private byte adjustChannel(byte cla) throws CardException {
        int newCla;
        if (channelNumber < 4) {
            newCla = (cla & 0xfc) | channelNumber;
        } else if (channelNumber < 16) {
            newCla = cla & 0x80; //keep bit 7
            newCla |= 0x40; //set bit 6
            newCla |= (cla & 0x1c) << 1; //shift bit 4-2 by 1
            newCla |= channelNumber;
        } else {
            throw new CardException("channel number too big: " + channelNumber);
        }
        return (byte) newCla;
    }

    /*-------------------------- Other Methods ------------------------------------------------------------*/

    /**
     * Request status reports
     * <p/>
     * <p>
     *     After calling this method the channel will send status reports indicating
     *     the total number of bytes send und received via this channel.
     * @param handler Handler to which the messages needs to be sent
     * @param what the status message what id to be used
     * </p>
     */
    public void requestStatusReports(Handler handler, int what){
        weakReferenceHandler = new WeakReference<Handler>(handler);
        this.what = what;
    }

    /**
     * Cancel status reports
     * <p/>
     * <p>
     *     After calling this method the channel will not send any status reports
     * </p>
     */
    public void cancelStatusReports(){
        weakReferenceHandler = null;
    }

}
