/*
 * SmartcardioNFC - a library providing a smartcardio implementation for NFC on Android
 *
 * Copyright (C) 2017 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH <info@gematik.de> +49 30 40041-0
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.nfcCard;

import android.app.Activity;
import android.nfc.NfcAdapter;
import android.nfc.Tag;
import android.nfc.tech.IsoDep;
import android.os.Bundle;
import android.util.Log;

import javax.smartcardio.CardException;
import javax.smartcardio.CardNotPresentException;
import javax.smartcardio.CardPermission;
import javax.smartcardio.CardTerminal;
import javax.smartcardio.TerminalFactory;

/**
 * A NFC Smart Card terminal, sometimes referred to as a NFC Card Reader.
 * A NfcCardTerminal object can be obtained by calling
 * {@linkplain NfcCardTerminals#list}
 * or {@linkplain NfcCardTerminals#getTerminal NfcCardTerminals.getTerminal()}.
 *
 * <p>Note that physical card readers with slots for multiple cards are
 * represented by one <code>CardTerminal</code> object per such slot.
 *
 * @see NfcCardTerminals
 * @see TerminalFactory
 */
public class NfcCardTerminal extends CardTerminal implements NfcAdapter.ReaderCallback {

    private static final String TAG = NfcCardTerminal.class.getName();
    private final String name;
    private NfcAdapter nfcAdapter;
    private NfcCard nfcCard = null;

    /**
     * Constructor.
     *
     * @param activity activity of an Android app
     */
    protected NfcCardTerminal(Activity activity) {
        this(activity, NfcCardTerminals.DEFAULT_NFC_ADAPTER);
    }

    /**
     * Constructor.
     *
     * @param activity activity of an Android app
     * @param name name
     */
    public NfcCardTerminal(Activity activity, String name) {
        this.name = name;

        Bundle options = null;
//        Bundle options = new Bundle();
//        options.putInt(NfcAdapter.EXTRA_READER_PRESENCE_CHECK_DELAY, 60000);

        nfcAdapter = NfcAdapter.getDefaultAdapter(activity);
        nfcAdapter.enableReaderMode(activity, this, NfcAdapter.FLAG_READER_NFC_A | NfcAdapter.FLAG_READER_NFC_B | NfcAdapter.FLAG_READER_SKIP_NDEF_CHECK, options);
    }

    /**
     * A callback to be invoked when the system finds a tag while the foreground activity is operating in reader mode.
     *
     * @param tag an NFC tag that has been discovered
     */
    public void onTagDiscovered(Tag tag) {
        Log.d(TAG, "onTagDiscovered(): " + tag.toString());
        Log.d(TAG, "NfcCardTerminal instance: " + this.toString());
        if (tag != null) {
            for(String tech: tag.getTechList()){
                if(tech.equals(IsoDep.class.getName())){
                    nfcCard = new NfcCard(tag);
                    break;
                }
            }
        }
    }

    /**
     * Returns the unique name of this terminal.
     *
     * @return the unique name of this terminal.
     */
    public String getName() {
        return name;
    }

    /**
     * Establishes a connection to the card.
     * If a connection has previously established using
     * the specified protocol, this method returns the same Card object as
     * the previous call.
     *
     * @param protocol the protocol to use ("T=0", "T=1", or "T=CL"), or "*" to
     *   connect using any available protocol.
     *
     * @throws NullPointerException if protocol is null
     * @throws IllegalArgumentException if protocol is an invalid protocol
     *   specification
     * @throws CardNotPresentException if no card is present in this terminal
     * @throws CardException if a connection could not be established
     *   using the specified protocol or if a connection has previously been
     *   established using a different protocol
     * @throws SecurityException if a SecurityManager exists and the
     *   caller does not have the required
     *   {@linkplain CardPermission permission}
     */
    public NfcCard connect(String protocol) throws CardException {
        if(!protocol.equals("T=CL") && !protocol.equals("T=*")) {
            throw new CardException("invalid protocol: " + protocol);
        }
        if (nfcCard == null) {
            throw new CardNotPresentException("no card");
        }
        nfcCard.connect();
        return nfcCard;
    }

    /**
     * Returns whether a card is present in this terminal.
     *
     * @return whether a card is present in this terminal.
     * @throws CardException if the status could not be determined
     */
    public boolean isCardPresent() {
        return nfcCard!=null&&nfcCard.isCardPresent();
    }

    /**
     * Waits until a card is present in this terminal or the timeout
     * expires. If the method returns due to an expired timeout, it returns
     * false. Otherwise it return true.
     * <p/>
     * <P>If a card is present in this terminal when this
     * method is called, it returns immediately.
     *
     * @param timeout if positive, block for up to <code>timeout</code>
     *                milliseconds; if zero, block indefinitely; must not be negative
     * @return false if the method returns due to an expired timeout,
     * true otherwise.
     * @throws IllegalArgumentException if timeout is negative
     * @throws CardException            if the operation failed
     */
    public boolean waitForCardPresent(long timeout) throws CardException {
        long start = System.currentTimeMillis();
        while (!isCardPresent()) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
            }
            if (System.currentTimeMillis() - start > timeout) {
                return false;
            }
        }
        return true;
    }

    /**
     * Waits until a card is absent in this terminal or the timeout
     * expires. If the method returns due to an expired timeout, it returns
     * false. Otherwise it return true.
     * <p/>
     * <P>If no card is present in this terminal when this
     * method is called, it returns immediately.
     * <p/>
     * @param timeout if positive, block for up to <code>timeout</code>
     *                milliseconds; if zero, block indefinitely; must not be negative
     * @return false if the method returns due to an expired timeout,
     * true otherwise.
     * @throws IllegalArgumentException if timeout is negative
     * @throws CardException            if the operation failed
     */
    public boolean waitForCardAbsent(long timeout) throws CardException {
        long start = System.currentTimeMillis();
        while (isCardPresent()) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
            }
            if (System.currentTimeMillis() - start > timeout) {
                return false;
            }
        }
        return true;
    }

}
