/*
 * SmartcardioNFC - a library providing a smartcardio implementation for NFC on Android
 *
 * Copyright (C) 2017 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH <info@gematik.de> +49 30 40041-0
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.nfcCard;

import android.app.Activity;
import android.nfc.NfcAdapter;
import android.util.Log;

import java.util.ArrayList;
import java.util.List;

import javax.smartcardio.CardException;
import javax.smartcardio.CardTerminals;

/**
 * List of NfcCardTerminals.
 */
public class NfcCardTerminals extends CardTerminals {

    private static final String TAG = NfcCardTerminals.class.getName();
    public final static String DEFAULT_NFC_ADAPTER = "DefaultNfcAdapter";
    private List<NfcCardTerminal> list = new ArrayList<>();
    State[] stateArray;
    private boolean obsolete = false;

    /**
     * Constructor.
     *
     * @param activity activity of an Android app
     */
    public NfcCardTerminals(Activity activity) {
        NfcAdapter nfcAdapter = NfcAdapter.getDefaultAdapter(activity);
        if (nfcAdapter == null) {
            Log.i(TAG, "no NFC adapter");
        } else {
            if (nfcAdapter.isEnabled()){
                list.add(new NfcCardTerminal(activity, DEFAULT_NFC_ADAPTER));
                Log.i(TAG, "NFC is enabled");
            }else{
                Log.i(TAG, "NFC is disabled");
            }
        }
        stateArray = new State[list.size()];
        int i=0;
        for(NfcCardTerminal nfcCardTerminal: list){
            stateArray[i++] = nfcCardTerminal.isCardPresent() ? State.CARD_PRESENT : State.CARD_ABSENT;
        }
    }

    /**
     * Mark this instance as obsolete.
     *
     * @param obsolete true if obsolete
     */
    public void setObsolete(boolean obsolete) {
        this.obsolete = obsolete;
    }

    /**
     * Waits for card insertion or removal in any of the terminals of this
     * object or until the timeout expires.
     *
     * <p>This method examines each CardTerminal of this object.
     * If a card was inserted into or removed from a CardTerminal since the
     * previous call to <code>waitForChange()</code>, it returns
     * immediately.
     * Otherwise, or if this is the first call to <code>waitForChange()</code>
     * on this object, it blocks until a card is inserted into or removed from
     * a CardTerminal.
     *
     * <p>If <code>timeout</code> is greater than 0, the method returns after
     * <code>timeout</code> milliseconds even if there is no change in state.
     * In that case, this method returns <code>false</code>; otherwise it
     * returns <code>true</code>.
     *
     * <p>This method is often used in a loop in combination with
     * {@link #list(CardTerminals.State) list(State.CARD_INSERTION)},
     * for example:
     * <pre>
     *  TerminalFactory factory = ...;
     *  CardTerminals terminals = factory.terminals();
     *  while (true) {
     *      for (CardTerminal terminal : terminals.list(CARD_INSERTION)) {
     *          // examine Card in terminal, return if it matches
     *      }
     *      terminals.waitForChange();
     *  }</pre>
     *
     * @param timeout if positive, block for up to <code>timeout</code>
     *   milliseconds; if zero, block indefinitely; must not be negative
     * @return false if the method returns due to an expired timeout,
     *   true otherwise.
     *
     * @throws IllegalStateException if this <code>CardTerminals</code>
     *   object does not contain any terminals
     * @throws IllegalArgumentException if timeout is negative
     * @throws CardException if the card operation failed
     */
    @Override
    public boolean waitForChange(long timeout) throws CardException {
        Log.d(TAG, "entered waitForChange()");
        int i = -1;
        boolean isChanged = false;
        long start = System.currentTimeMillis();
        while(!isChanged && !Thread.currentThread().isInterrupted() && !obsolete){
            if (i != -1) {  // do not already sleep right at the beginning
                try {
//                    Log.d(TAG, "wait 100 ms");
                    Thread.sleep(100);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
                if (timeout!=0 && System.currentTimeMillis() - start > timeout) {
                    Log.d(TAG, "left waitForChange() with timeout");
                    return false;
                }
            }

            i=0;
            for(NfcCardTerminal nfcCardTerminal: list){
//                Log.d(TAG, "NfcCardTerminal instance: " + nfcCardTerminal.toString());
                if(nfcCardTerminal.isCardPresent()) {
//                    Log.d(TAG, "isCardPresent = true");
                    switch (stateArray[i]) {
                        case CARD_ABSENT:
                            Log.d(TAG, "case CARD_ABSENT -> CARD_INSERTION");
                            stateArray[i] = State.CARD_INSERTION;
                            isChanged = true;
                            break;
                        case CARD_INSERTION:
                            Log.d(TAG, "case CARD_INSERTION -> CARD_PRESENT");
                            stateArray[i] = State.CARD_PRESENT;
                            break;
                        case CARD_REMOVAL:
                            Log.d(TAG, "case CARD_REMOVAL -> CARD_INSERTION");
                            stateArray[i] = State.CARD_INSERTION;
                            isChanged = true;
                            break;
                        case CARD_PRESENT:
//                            Log.d(TAG, "case CARD_PRESENT -> CARD_PRESENT");
                            break;
                    }
                } else {
//                    Log.d(TAG, "isCardPresent = false");
                    switch (stateArray[i]) {
                        case CARD_ABSENT:
//                            Log.d(TAG, "case CARD_ABSENT -> CARD_ABSENT");
                            break;
                        case CARD_INSERTION:
                            Log.d(TAG, "case CARD_INSERTION -> CARD_REMOVAL");
                            stateArray[i] = State.CARD_REMOVAL;
                            isChanged = true;
                            break;
                        case CARD_REMOVAL:
                            Log.d(TAG, "case CARD_REMOVAL -> CARD_ABSENT");
                            stateArray[i] = State.CARD_ABSENT;
                            break;
                        case CARD_PRESENT:
                            Log.d(TAG, "case CARD_PRESENT -> CARD_REMOVAL");
                            stateArray[i] = State.CARD_REMOVAL;
                            isChanged = true;
                            break;
                    }
                }
                i++;
            }
        }

        if (obsolete) {
            Log.d(TAG, "left waitForChange() with obsolete");
            return false;
        } else if (isChanged){
            Log.d(TAG, "left waitForChange() with change");
            return true;
        } else {
            Log.d(TAG, "left waitForChange() with interrupt");
            return true;
        }
    }

    /**
     * Returns an unmodifiable list of all terminals matching the specified
     * stateArray.
     *
     * <p>If stateArray is {@link State#ALL State.ALL}, this method returns
     * all CardTerminals encapsulated by this object.
     * If stateArray is {@link State#CARD_PRESENT State.CARD_PRESENT} or
     * {@link State#CARD_ABSENT State.CARD_ABSENT}, it returns all
     * CardTerminals where a card is currently present or absent, respectively.
     *
     * <p>If stateArray is {@link State#CARD_INSERTION State.CARD_INSERTION} or
     * {@link State#CARD_REMOVAL State.CARD_REMOVAL}, it returns all
     * CardTerminals for which an insertion (or removal, respectively)
     * was detected during the last call to {@linkplain #waitForChange}.
     * If <code>waitForChange()</code> has not been called on this object,
     * <code>CARD_INSERTION</code> is equivalent to <code>CARD_PRESENT</code>
     * and <code>CARD_REMOVAL</code> is equivalent to <code>CARD_ABSENT</code>.
     * For an example of the use of <code>CARD_INSERTION</code>,
     * see {@link #waitForChange}.
     *
     * @param state the State
     * @return an unmodifiable list of all terminals matching the specified
     *   attribute.
     *
     * @throws NullPointerException if attr is null
     * @throws CardException if the card operation failed
     */
    @Override
    public List<NfcCardTerminal> list(State state) throws CardException {
        List<NfcCardTerminal> cardTerminalList = new ArrayList<>();
        int i = 0;
        for (NfcCardTerminal cardTerminal: list){
            if(state==State.ALL||stateArray[i]==state) {
                cardTerminalList.add(cardTerminal);
            }
            i++;        }
        return cardTerminalList;
    }
}
